import ast
from typing import List, Optional

from pydantic_settings import SettingsConfigDict
from turms.config import GeneratorConfig
from graphql import GraphQLSchema, InputObjectTypeDefinitionNode
from graphql.language.ast import OperationDefinitionNode, OperationType
from turms.recurse import type_field_node
from turms.plugins.base import Plugin, PluginConfig
from pydantic import Field
from graphql.language.ast import (
    FieldNode,
    OperationDefinitionNode,
    OperationType,
)
from graphql.utilities.get_operation_root_type import get_operation_root_type
from graphql.utilities.type_info import get_field_def

import re
from graphql import NonNullTypeNode, VariableDefinitionNode, language
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    inspect_operation_for_documentation,
    merge_bases_sequences,
    merge_body_sequences,
    parse_documents,
    recurse_type_annotation,
    replace_iteratively,
    parse_value_node,
)
import logging


logger = logging.getLogger(__name__)
fragment_searcher = re.compile(r"\.\.\.(?P<fragment>[a-zA-Z]*)")


class OperationsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_OPERATIONS_")
    type: str = "turms.plugins.operations.OperationsPlugin"
    query_bases: List[str] = []
    arguments_bases: List[str] = []
    mutation_bases: List[str] = []
    subscription_bases: List[str] = []
    operations_glob: Optional[str] = None
    create_arguments: bool = True
    extract_documentation: bool = True
    arguments_allow_population_by_field_name: bool = False


def get_query_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.query_bases:
        for base in plugin_config.query_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.query_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def get_mutation_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.mutation_bases:
        for base in plugin_config.mutation_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.mutation_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def generate_arguments_config(
    o: OperationType,
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if config.pydantic_version == "1":
        config_fields = []

        if plugin_config.arguments_allow_population_by_field_name:
            config_fields.append(
                ast.Assign(
                    targets=[
                        ast.Name(id="allow_population_by_field_name", ctx=ast.Store())
                    ],
                    value=ast.Constant(value=True),
                )
            )

        if len(config_fields) > 0:
            return [
                ast.ClassDef(
                    name="Config",
                    bases=[],
                    keywords=[],
                    body=[
                        ast.Expr(
                            value=ast.Constant(value=" Arguments config class"),
                        )
                    ]
                    + config_fields,
                    decorator_list=[],
                )
            ]
        else:
            return []

    else:
        config_keywords = []

        if plugin_config.arguments_allow_population_by_field_name is not None:
            config_keywords.append(
                ast.keyword(
                    arg="populate_by_name",
                    value=ast.Constant(
                        value=config.options.allow_population_by_field_name
                    ),
                )
            )

        if len(config_keywords) > 0:
            registry.register_import("pydantic.ConfigDict")
            return [
                ast.Assign(
                    targets=[ast.Name(id="model_config", ctx=ast.Store())],
                    value=ast.Call(
                        func=ast.Name(id="ConfigDict", ctx=ast.Load()),
                        args=[],
                        keywords=config_keywords,
                    ),
                )
            ]
        else:
            return []


def get_arguments_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
) -> List[ast.Name]:
    if plugin_config.arguments_bases:
        for base in plugin_config.arguments_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.arguments_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def get_subscription_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.subscription_bases:
        for base in plugin_config.subscription_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.subscription_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def generate_operation(
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    tree = []
    if o.name is None:
        raise ValueError(
            "Operation name is required. Please provide a name for the operation."
        )

    # Generation means creating a class for the operation
    if o.operation == OperationType.MUTATION:
        class_name = registry.generate_mutation(o.name.value)
        extra_bases = get_mutation_bases(config, plugin_config, registry)
    elif o.operation == OperationType.SUBSCRIPTION:
        class_name = registry.generate_subscription(o.name.value)
        extra_bases = get_subscription_bases(config, plugin_config, registry)
    elif o.operation == OperationType.QUERY:
        class_name = registry.generate_query(o.name.value)
        extra_bases = get_query_bases(config, plugin_config, registry)
    else:
        raise ValueError(
            f"Unsupported operation type: {o.operation}. Supported types are: QUERY, MUTATION, SUBSCRIPTION."
        )

    x = get_operation_root_type(client_schema, o)
    class_body_fields = []

    operation_documentation = (
        inspect_operation_for_documentation(o)
        if plugin_config.extract_documentation
        else None
    )
    if operation_documentation:
        class_body_fields.append(
            ast.Expr(
                value=ast.Constant(value=operation_documentation),
            )
        )
    else:
        class_body_fields.append(
            ast.Expr(
                value=ast.Constant(value="No documentation found for this operation."),
            )
        )

    operation_annotations: list[ast.AnnAssign] = []

    for field_node in o.selection_set.selections:
        if isinstance(field_node, FieldNode):
            field_definition = get_field_def(client_schema, x, field_node)
            assert field_definition, "Couldn't find field definition"

            generated_node = type_field_node(
                field_node,
                class_name,
                field_definition,
                client_schema,
                config,
                tree,
                registry,
            )

            type_annotation = generated_node[0]
            if isinstance(type_annotation, ast.AnnAssign):
                operation_annotations.append(type_annotation)
            else:
                raise ValueError(
                    f"Expected type annotation to be of type ast.AnnAssign, got {type(type_annotation)}"
                )

            class_body_fields += generated_node

        else:
            raise ValueError(f"Unsupported selection set node type: {type(field_node)}")

    if len(operation_annotations) == 1:
        # "Has only one root operation -> Registering as a single operation"
        registry.register_operation_single(
            o.name.value, operation_annotations[0].annotation
        )

    query_document = language.print_ast(o)
    merged_document = replace_iteratively(query_document, registry)

    if plugin_config.create_arguments:
        arguments_body = []

        for v in o.variable_definitions:
            is_optional = not isinstance(v.type, NonNullTypeNode) or v.default_value
            annotation = recurse_type_annotation(v.type, registry)
            field_name = registry.generate_parameter_name(v.variable.name.value)
            target = v.variable.name.value

            if target != field_name:
                registry.register_import("pydantic.Field")
                if is_optional:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="alias", value=ast.Constant(value=target)
                                ),
                                ast.keyword(
                                    arg="default",
                                    value=ast.Constant(
                                        value=(
                                            parse_value_node(v.default_value)
                                            if v.default_value is not None
                                            else None
                                        )
                                    ),
                                ),
                            ],
                        ),
                        simple=1,
                    )
                else:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="alias", value=ast.Constant(value=target)
                                )
                            ],
                        ),
                        simple=1,
                    )
            else:
                if is_optional:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="default",
                                    value=ast.Constant(
                                        value=(
                                            parse_value_node(v.default_value)
                                            if v.default_value is not None
                                            else None
                                        )
                                    ),
                                ),
                            ],
                        ),
                        simple=1,
                    )
                else:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        simple=1,
                    )

            arguments_body += [assign]

        arguments_body += generate_arguments_config(
            o.operation, config, plugin_config, registry
        )

        class_body_fields += [
            ast.ClassDef(
                "Arguments",
                bases=merge_bases_sequences(
                    get_arguments_bases(config, plugin_config, registry=registry)
                ),
                decorator_list=[],
                keywords=[],
                type_params=[],
                body=merge_body_sequences(
                    [
                        ast.Expr(
                            value=ast.Constant(value=f"Arguments for {o.name.value} "),
                        )
                    ],
                    arguments_body or [ast.Pass()],
                ),
            )
        ]

    meta_body = [
        ast.Expr(
            value=ast.Constant(value=f"Meta class for {o.name.value} "),
        ),
        ast.Assign(
            targets=[ast.Name(id="document", ctx=ast.Store())],
            value=ast.Constant(value=merged_document),
        ),
    ]
    if config.domain:
        meta_body += [
            ast.Assign(
                targets=[ast.Name(id="domain", ctx=ast.Store())],
                value=ast.Constant(value=str(config.domain)),
            )
        ]

    class_body_fields += [
        ast.ClassDef(
            "Meta",
            bases=[],
            decorator_list=[],
            keywords=[],
            body=meta_body,
            type_params=[],
        ),
    ]

    tree.append(
        ast.ClassDef(
            class_name,
            bases=merge_bases_sequences(extra_bases),
            decorator_list=[],
            keywords=[],
            type_params=[],
            body=merge_body_sequences(
                class_body_fields,
                generate_pydantic_config(o.operation, config, registry),
            ),
        )
    )

    return tree


class OperationsPlugin(Plugin):
    """ " Generate operations as classes

    This plugin created classes for operations. It will scan your documents and create classes for each operation.
    The class will have a `document` attribute that contains the query document, as well as contained "Arguments" class
    with the variables for the operation.

    This allows for the serialization of values in both directions.

    If you want to generate python functions instead, use the `funcs` plugin in ADDITION to this plugin.

    """

    config: OperationsPluginConfig = Field(default_factory=OperationsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        plugin_tree = []

        glob = self.config.operations_glob or config.documents

        if not glob:
            raise ValueError(
                "No documents found. Please provide a glob pattern for the documents."
            )

        documents = parse_documents(client_schema, glob)

        definitions = documents.definitions
        operations = [
            node for node in definitions if isinstance(node, OperationDefinitionNode)
        ]

        for operation in operations:
            plugin_tree += generate_operation(
                operation, client_schema, config, self.config, registry
            )

        return plugin_tree



from dataclasses import dataclass
from graphql import (
    ConstArgumentNode,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
)
from pydantic_settings import SettingsConfigDict
from turms.plugins.base import Plugin, PluginConfig
import ast
from typing import Dict, List, Optional
from turms.config import GeneratorConfig
from graphql import GraphQLSchema
from turms.plugins.base import Plugin
from pydantic import Field
from graphql.type.definition import (
    GraphQLEnumType,
)
from turms.referencer import create_reference_registry_from_documents
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    get_additional_bases_for_type,
    parse_documents,
)
from turms.config import GraphQLTypes


class InputsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(
        extra="forbid", env_prefix="TURMS_PLUGINS_INPUTS_"
    )
    type: str = "turms.plugins.inputs.InputsPlugin"
    inputtype_bases: List[str] = ["pydantic.BaseModel"]
    allow_population_by_field_name: bool = True
    skip_underscore: bool = True
    skip_unreferenced: bool = True


def generate_input_annotation(
    type: GraphQLScalarType | GraphQLEnumType | GraphQLInputObjectType,
    parent: str,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(type, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(type.name),
            )

        return registry.reference_scalar(type.name)

    if isinstance(type, GraphQLInputObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_inputtype(type.name, parent),
                ctx=ast.Load(),
            )
        return registry.reference_inputtype(type.name, parent)

    if isinstance(type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(
                    type.name, parent, allow_forward=not config.force_plugin_order
                ),
                ctx=ast.Load(),
            )
        return registry.reference_enum(
            type.name, parent, allow_forward=not config.force_plugin_order
        )

    if isinstance(type, GraphQLNonNull):
        return generate_input_annotation(
            type.of_type, parent, config, plugin_config, registry, is_optional=False
        )

    if isinstance(type, GraphQLList):
        if (
            config.freeze.enabled
            and GraphQLTypes.INPUT in config.freeze.types
            and config.freeze.convert_list_to_tuple
        ):
            registry.register_import("typing.Tuple")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("Tuple", ctx=ast.Load()),
                    slice=ast.Tuple(elts=[x, ast.Constant(value=...)], ctx=ast.Load()),
                    ctx=ast.Load(),
                )

        else:
            registry.register_import("typing.List")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()), slice=x, ctx=ast.Load()
                )

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    generate_input_annotation(
                        type.of_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    )
                ),
                ctx=ast.Load(),
            )

        return list_builder(
            generate_input_annotation(
                type.of_type, parent, config, plugin_config, registry, is_optional=True
            )
        )

    raise NotImplementedError(f"Unknown input type {type}")


@dataclass
class Discriminator:
    discriminator: str
    value: str


def generate_input_type(
    name: str,
    union_type_descriminators: Dict[str, str],
    type: GraphQLInputType,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
    key: str,
    discriminator: Optional[Discriminator] = None,
):
    additional_bases = get_additional_bases_for_type(type.name, config, registry)

    fields = (
        [ast.Expr(value=ast.Constant(value=type.description))]
        if type.description
        else [ast.Expr(value=ast.Constant(value="No documentation"))]
    )

    if discriminator:
        fields.append(
            ast.AnnAssign(
                target=ast.Name(discriminator.discriminator, ctx=ast.Store()),
                annotation=ast.Subscript(
                    value=ast.Name("Literal", ctx=ast.Load()),
                    slice=ast.Constant(value=discriminator.value),
                    ctx=ast.Load(),
                ),
                value=ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=[
                        ast.keyword(
                            arg="default", value=ast.Constant(value=discriminator.value)
                        )
                    ],
                ),
                simple=1,
            )
        )

    for value_key, value in type.fields.items():
        field_name = registry.generate_node_name(value_key)

        if field_name != value_key:
            registry.register_import("pydantic.Field")

            keywords = [ast.keyword(arg="alias", value=ast.Constant(value=value_key))]
            if not isinstance(value.type, GraphQLNonNull):
                keywords.append(ast.keyword(arg="default", value=ast.Constant(None)))

            assign = ast.AnnAssign(
                target=ast.Name(field_name, ctx=ast.Store()),
                annotation=generate_input_annotation(
                    value.type,
                    name,
                    config,
                    plugin_config,
                    registry,
                    is_optional=True,
                ),
                value=ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=keywords,
                ),
                simple=1,
            )

        else:
            assign = ast.AnnAssign(
                target=ast.Name(value_key, ctx=ast.Store()),
                annotation=generate_input_annotation(
                    value.type,
                    name,
                    config,
                    plugin_config,
                    registry,
                    is_optional=True,
                ),
                simple=1,
                value=(
                    ast.Constant(None)
                    if not isinstance(value.type, GraphQLNonNull)
                    else None
                ),
            )

        potential_comment = (
            value.description
            if not value.deprecation_reason
            else f"DEPRECATED: {value.description}"
        )

        if potential_comment:
            fields += [
                assign,
                ast.Expr(value=ast.Constant(value=potential_comment)),
            ]

        else:
            fields += [assign]

    return ast.ClassDef(
        name,
        bases=additional_bases
        + [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.inputtype_bases
        ],
        decorator_list=[],
        keywords=[],
        body=fields
        + generate_pydantic_config(GraphQLTypes.INPUT, config, registry, typename=key),
    )


def generate_inputs(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    inputobjects_type = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLInputObjectType)
    }

    if plugin_config.skip_unreferenced and config.documents:
        ref_registry = create_reference_registry_from_documents(
            client_schema, parse_documents(client_schema, config.documents)
        )
    else:
        ref_registry = None

    for base in plugin_config.inputtype_bases:
        registry.register_import(base)

    union_input_types = {}
    union_type_discriminators = {}

    for key, type in inputobjects_type.items():
        directives = type.ast_node.directives if type.ast_node else []
        for directive in directives:
            directive_name = directive.name.value
            if directive_name == "unionElementOf":
                union_type = None
                discriminator = None
                key = None
                for arg in directive.arguments:
                    if isinstance(arg.value, ConstArgumentNode):
                        if arg.name.value == "union":
                            union_type = arg.value.value
                        if arg.name.value == "discriminator":
                            discriminator = arg.value.value
                        if arg.name.value == "key":
                            key = arg.value.value

                if union_type in ref_registry.inputs:
                    if union_type not in union_input_types:
                        union_input_types[union_type] = []
                    if union_type not in union_type_discriminators:
                        union_type_discriminators[union_type] = discriminator

                    assert union_type_discriminators[union_type] == discriminator, (
                        f"Discriminator mismatch for {union_type} expected {union_type_discriminators[union_type]} got {discriminator}"
                    )

                    name = registry.generate_inputtype(type.name)
                    union_input_types[union_type].append(name)
                    tree.append(
                        generate_input_type(
                            name,
                            union_type_discriminators,
                            type,
                            config,
                            plugin_config,
                            registry,
                            type.name,
                            Discriminator(discriminator=discriminator, value=key),
                        )
                    )

    for key, type in inputobjects_type.items():
        if ref_registry and key not in ref_registry.inputs:
            continue

        if plugin_config.skip_underscore and key.startswith("_"):  # pragma: no cover
            continue

        if type.name in union_input_types:
            registry.register_import("typing.Union")
            registry.register_import("typing.Annotated")
            registry.register_import("pydantic.Field")
            union_slice = ast.Tuple(
                elts=[
                    ast.Name(id=clsname, ctx=ast.Load())
                    for clsname in union_input_types[type.name]
                ],
                ctx=ast.Load(),
            )

            slice = ast.Tuple(
                elts=[
                    ast.Subscript(
                        value=ast.Name("Union", ctx=ast.Load()),
                        slice=union_slice,
                        ctx=ast.Load(),
                    ),
                    ast.Call(
                        func=ast.Name(id="Field", ctx=ast.Load()),
                        args=[],
                        keywords=[
                            ast.keyword(
                                arg="discriminator",
                                value=ast.Constant(
                                    union_type_discriminators[type.name]
                                ),
                            )
                        ],
                    ),
                ],
                ctx=ast.Load(),
            )

            tree.append(
                ast.Assign(
                    targets=[
                        ast.Name(
                            id=registry.generate_inputtype(type.name),
                            ctx=ast.Store(),
                        )
                    ],
                    value=ast.Subscript(
                        value=ast.Name("Annotated", ctx=ast.Load()),
                        slice=slice,
                        ctx=ast.Load(),
                    ),
                )
            )

            continue

        additional_bases = get_additional_bases_for_type(type.name, config, registry)
        name = registry.generate_inputtype(key)
        fields = (
            [ast.Expr(value=ast.Constant(value=type.description))]
            if type.description
            else [ast.Expr(value=ast.Constant(value="No documentation"))]
        )

        for value_key, value in type.fields.items():
            field_name = registry.generate_node_name(value_key)

            if field_name != value_key:
                registry.register_import("pydantic.Field")
                keywords = [
                    ast.keyword(arg="alias", value=ast.Constant(value=value_key))
                ]
                if not isinstance(value.type, GraphQLNonNull):
                    keywords.append(
                        ast.keyword(arg="default", value=ast.Constant(None))
                    )

                assign = ast.AnnAssign(
                    target=ast.Name(field_name, ctx=ast.Store()),
                    annotation=generate_input_annotation(
                        value.type,
                        name,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    ),
                    value=ast.Call(
                        func=ast.Name(id="Field", ctx=ast.Load()),
                        args=[],
                        keywords=keywords,
                    ),
                    simple=1,
                )

            else:
                assign = ast.AnnAssign(
                    target=ast.Name(value_key, ctx=ast.Store()),
                    annotation=generate_input_annotation(
                        value.type,
                        name,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    ),
                    simple=1,
                    value=(
                        ast.Constant(None)
                        if not isinstance(value.type, GraphQLNonNull)
                        else None
                    ),
                )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        tree.append(
            ast.ClassDef(
                name,
                bases=additional_bases
                + [
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.inputtype_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields
                + generate_pydantic_config(
                    GraphQLTypes.INPUT, config, registry, typename=key
                ),
            )
        )

    return tree


class InputsPlugin(Plugin):
    """Generate pydantic Models for GraphQL inputs

    This plugin generates pydantic models for all GraphQL inputtypes in
    your schema. It will generate a model for each inputtype in your schema.

    """

    config: InputsPluginConfig = Field(default_factory=InputsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        for base in self.config.inputtype_bases:
            registry.register_import(base)

        return generate_inputs(client_schema, config, self.config, registry)
